"""
Files

This module contains actions used to read & write files in the task workspace.

The task workspace contains any supporting playbook files, & any files created by other plugins.

!!! danger "PPA Version"

    The actions in this module require PPA v2.7.x or newer.
"""
import os
import json
import logging
import yaml

from typing import List, Dict, Any, Union

from ppa import ui


logger = logging.getLogger(__name__)

AnyJson = Union[None, bool, int, float, str, List[Any], Dict[str, Any]]

SHARED_WORKSPACE = "/workspace/share"


def read_text_file(name: str) -> str:
    """Read a text file from the task workspace.

    !!! question "Minimum Plugin Version: 4.0.0"
    !!! question "Minimum PPA Version: 2.7.0"

    ??? info "Input"

        - __name:__ the name of the file including the extension

    ??? success "Output"

        The file contents as a string.

    ??? example
        ```YAML
        - ppa_tools.files.read_text_file:
            name: intro.md
          save: introduction

        - ppa.ui.output_markdown:
            doc: "{{ introduction }}"
        ```
    """
    file_path = os.path.join(SHARED_WORKSPACE, name)

    if not os.path.isfile(file_path):
        message = f"Couldn't read contents from file '{name}' as it does not exist in the task workspace."
        ui.output_error(message)
        raise ValueError(message)

    logger.info(f"reading contents from {file_path}")
    with open(file_path, "rb") as open_file:
        contents = open_file.read().decode("utf-8")
    return contents


def read_json_file(name: str) -> AnyJson:
    """Read a JSON file from the task workspace.

    !!! question "Minimum Plugin Version: 4.0.0"

    !!! question "Minimum PPA Version: 2.7.0"

    ??? info "Input"

        - __name:__ the name of the file including the extension

    ??? success "Output"

        A loaded JSON object.

    ??? example
        ```YAML
        - ppa_tools.files.read_json_file:
            name: config.json
          save: config
        ```
    """
    try:
        return json.loads(read_text_file(name))
    except json.JSONDecodeError as e:
        logger.error(e)
        message = f"Failed to load contents of file '{name}' from the task workspace as it is not valid JSON."
        ui.output_error(message)
        raise ValueError(message)


def read_yaml_file(name: str) -> AnyJson:
    """Read a YAML file from the task workspace.

    !!! question "Minimum Plugin Version: 5.8.0"

    !!! question "Minimum PPA Version: 2.7.0"

    ??? info "Input"

        - __name:__ the name of the file including the extension

    ??? success "Output"

        A loaded YAML object.

    ??? example
        ```YAML
        - ppa_tools.files.read_yaml_file:
            name: config.yaml
          save: config
        ```
    """
    try:
        return yaml.safe_load(read_text_file(name))
    except yaml.scanner.ScannerError as e:
        logger.error(e)
        message = f"Failed to load contents of file '{name}' from the task workspace as it is not valid YAML."
        ui.output_error(message)
        raise ValueError(message)


def write_text_file(name: str, contents: str) -> None:
    """Write a text file to the task workspace.

    !!! question "Minimum Plugin Version: 4.0.0"
    !!! question "Minimum PPA Version: 2.7.0"

    ??? info "Input"

        - __name:__ the name of the file including the extension

        - __contents:__ the file contents to write

    ??? success "Output"

        Nothing is outputted by this action.

    ??? example
        ```YAML
        - ppa.ui.input_text:
            text: Supply some text
          save: text

        - ppa_tools.files.write_text_file:
            name: example.txt
          load:
            contents: text
        ```
    """
    file_path = os.path.join(SHARED_WORKSPACE, name)

    if any([os.path.isfile(file_path), os.path.isdir(file_path)]):
        message = f"The supplied file name '{name}' already exists as a file or directory."
        ui.output_error(message)
        raise ValueError(message)

    logger.info(f"writing contents to {file_path}")
    with open(file_path, "w+") as new_file:
        new_file.write(contents)


def write_json_file(name: str, contents: AnyJson) -> None:
    """Write a JSON file to the task workspace.

    This action will fail if the supplied `items` cannot be converted to JSON.

    !!! question "Minimum Plugin Version: 4.0.0"
    !!! question "Minimum PPA Version: 2.7.0"

    ??? info "Input"

        - __name:__ the name of the file including the extension

        - __contents:__ any JSON serializable value

    ??? success "Output"

        Nothing is outputted by this action.

    ??? example
        ```YAML
        - set:
            name: example_json
            value:
              key_1: value_1
              key_2: value_2

        - ppa_tools.files.write_json_file:
            name: example.json
          load:
            contents: example_json
        ```
    """
    try:
        json_string = json.dumps(contents)
    except Exception as e:
        logger.error(e)
        message = f"Failed to write contents to file '{name}' as the contents cannot be converted to JSON."
        ui.output_error(message)
        raise ValueError(message)

    return write_text_file(name, json_string)
