"""
Lists

This module contains actions used to process lists.
"""
import logging
import re

from typing import List, Any, Optional, Dict, Union

from typeguard import check_argument_types

from ppa import ui


logger = logging.getLogger(__name__)


AnyJson = Union[None, bool, int, float, str, List[Any], Dict[str, Any]]


def _get_item(items, index):
    try:
        return items[index]
    except IndexError:
        message = f"The supplied list contains no values at index {index}."
        ui.output_error(message)
        raise IndexError(message)


def filter(items: List[AnyJson], value: AnyJson) -> List[AnyJson]:
    """Filter items from a list that match the supplied value.

    !!! question "Minimum Plugin Version: 6.3.0"

    ??? info "Input"

        - __items:__ the list of values to filter

        - __value:__ the value to filter from the list

    ??? success "Output"

        A new list with all occurrences of **value** removed.

    ??? example

        Filtering out `null` values from a list:

        ```YAML linenums="1"
        - ppa_tools.lists.filter:
            items:
              - null
              - 1
              - 2
            value: null
          save: filtered_list
        ```

        In this example the `filtered_list` variable will be:

        ```YAML linenums="1"
        - 1
        - 2
        ```
    """
    if value in {None, True, False}:
        operator = lambda item: item is not value
    else:
        operator = lambda item: item != value
    return [i for i in items if operator(i) is True]


def sort(items: List[Any], reverse: Optional[bool] = False) -> List[Any]:
    """Sort a list alphanumerically.

    !!! question "Minimum Plugin Version: 3.7.0"

    ??? info "Input"

        - __items:__ the list to sort

        - __reverse:__ set to `true` to sort in reverse (defaults to `false`)

    !!! danger "Value Types"

        In order to sort a list, all the items must be of the same type.

        Lists that contain `null` values cannot be sorted.

    ??? success "Output"

        A new sorted list.

    ??? example

        Sorting a list of numbers:

        ```YAML linenums="1"
        - ppa_tools.lists.sort:
            items:
              - 5
              - 4
              - 3
              - 2
              - 1
          save: sorted_list
        ```

        In this example the `sorted_list` variable will be:

        ```YAML linenums="1"
        - 1
        - 2
        - 3
        - 4
        - 5
        ```
    """
    if any([i is None for i in items]):
        logger.error(f"items: {items}")
        message = "Cannot sort the list as one or more of the values is null."
        ui.output_error(message)
        raise TypeError(message)

    types = list(set(map(lambda i: type(i), items)))

    if len(types) != 1:
        logger.error(f"items: {items}")
        regex = re.compile(r"<class\s'(.*)'>")
        type_names = ", ".join(list(map(lambda t: re.search(regex, t).group(1), [str(t) for t in types])))
        message = f"In order to sort a list all the items must be the same type. Received '{type_names}'."
        ui.output_error(message)
        raise TypeError(message)

    return sorted(items, reverse=reverse)


def common(first: List[Any], second: List[Any]) -> Any:
    """Compare two lists and get all items present in both.

    !!! question "Minimum Plugin Version: 3.5.0"

    ??? info "Input"

        - __first:__ the first list

        - __second:__ the second list

    ??? success "Output"

        A new list containing all items present in both `first` _and_ `second`.

    ??? example

        Getting the common items from two lists of numbers:

        ```YAML linenums="1"

        - ppa_tools.lists.common:
            first:
              - 1
              - 2
              - 3
              - 4
              - 5
            second:
              - 3
              - 4
              - 5
              - 6
              - 7
          save: new_list
        ```

        In this example the `new_list` variable will be:

        ```YAML linenums="1"
        - 3
        - 4
        - 5
        ```
    """
    try:
        check_argument_types()
    except TypeError:
        message = "Both inputs to the common action must be lists."
        ui.output_error(message)
        raise TypeError(message)
    return [x for x in first if x in second]


def unique(items: List[Any]) -> Any:
    """Get a list of unique values from the supplied list.

    This action can be used to remove duplicate values from a list.

    !!! question "Minimum Plugin Version: 3.5.0"

    ??? info "Input"

        - __items:__ a list of values

    ??? danger "Supported Values"

        This action will fail if the list contains a dictionary value.

    ??? success "Output"

        A new list containing all unique values from `items`.

    ??? example

        ```YAML linenums="1"
        - ppa_tools.lists.unique:
            items:
              - 1
              - 1
              - 2
              - 3
              - 3
              - 3
          save: unique_values
        ```

        In this example the `unique_values` variable will be:

        ```YAML linenums="1"
        - 1
        - 2
        - 3
        ```
    """
    if any([isinstance(v, dict) for v in items]):
        message = "Dictionary values are not supported by this action."
        ui.output_error(message)
        raise TypeError(message)

    return list(set(items))


def difference(first: List[Any], second: List[Any], key: Optional[str] = None) -> List[Any]:
    """Compare two lists and get all items only present in the first.

    !!! question "Minimum Plugin Version: 5.5.0"

    ??? info "Input"

        - __first:__ the first list

        - __second:__ the second list

        - __key:__ optional key to use in comparison checks if _both lists are dictionaries_

    ??? success "Output"

        A new list containing all items from `first` that are _not_ present in `second`.

    ??? example "Example 1 - Lists of Numbers"

        Getting the difference between two lists of numbers:

        ```YAML linenums="1"

        - ppa_tools.lists.difference:
            first:
              - 1
              - 2
              - 3
            second:
              - 3
              - 4
              - 5
          save: new_list
        ```

        In this example the `new_list` variable will be:

        ```
        - 1
        - 2
        ```

    ??? example "Example 2 - Lists of Dictionaries"

        - Getting the difference between two lists of Active Directory users

        - Each user's _sAMAccountName_ is used to check for existence

        ```YAML linenums="1"
        - active_directory.groups.get_users:
            distinguishedName: CN=Remote Desktop Users,CN=Builtin,DC=Example,DC=Domain
          load:
            domain_controller: domain_controller
          save: remote_desktop_users

        - active_directory.groups.get_users:
            distinguishedName: CN=Backup Operators,CN=Builtin,DC=Example,DC=Domain
          load:
            domain_controller: domain_controller
          save: backup_operators

        - ppa_tools.lists.difference:
            key: sAMAccountName
          load:
            first: backup_operators
            second: remote_desktop_users
          save: backup_operators_only
        ```
    """
    try:
        check_argument_types()
    except TypeError:
        message = "Both inputs to the difference action must be lists."
        ui.output_error(message)
        raise TypeError(message)

    if key is not None:

        if not all((isinstance(i, dict) for i in first + second)):
            message = "The supplied lists can only contain dictionaries when the 'key' input is supplied."
            ui.output_error(message)
            raise TypeError(message)

        if not all((key in d for d in (first + second))):
            message = f"Key '{key}' does not exist in one of the supplied dictionaries."
            ui.output_error(message)
            raise KeyError(message)

        first_only = []
        second_key_values = [s[key] for s in second]
        for item in first:
            if item[key] not in second_key_values:
                first_only.append(item)
        return first_only

    return [x for x in first if x not in second]


def combine(first: List[Any], second: List[Any]) -> Any:
    """Combine two lists into one.

    !!! question "Minimum Plugin Version: 3.5.0"

    ??? info "Input"

        - __first:__ the first list

        - __second:__ the second list

    ??? success "Output"

        A new list containing `first` & `second`.

    ??? example

        Combining two lists of numbers:

        ```YAML linenums="1"

        - ppa_tools.lists.combine:
            first:
              - 1
              - 2
            second:
              - 3
              - 4
          save: new_list
        ```

        In this example the `new_list` variable will be:

        ```YAML linenums="1"
        - 1
        - 2
        - 3
        - 4
        ```
    """
    try:
        check_argument_types()
    except TypeError:
        message = "Both inputs to the combine action must be lists."
        ui.output_error(message)
        raise TypeError(message)
    return first + second


def append(items: List[Any], value: Any) -> Any:
    """Append an item to a list.

    !!! question "Minimum Plugin Version: 3.1.0"

    ??? info "Input"

        - __items:__ the list to append the new `value` to

        - __value:__ the value to append to `items`

    ??? success "Output"

        The updated list.

    ??? example

        Appending an integer to a list.

        ```YAML linenums="1"
        - ppa_tools.lists.append:
            items:
              - 1
              - 2
            value: 3
          save: numbers
        ```

        The `numbers` variable will be:

        ```YAML linenums="1"
        - 1
        - 2
        - 3
        ```
    """
    items.append(value)
    return items


def get_first_item(items: List[Any]) -> Any:
    """Get the first item from a list.

    !!! question "Minimum Plugin Version: 3.0.0"

    ??? info "Input"

        - __items:__ a list of values

    ??? success "Output"

        The value at position `0` in the `items` list.

    ??? example

        Getting the first item from a list of cities.

        ```YAML linenums="1"
        - ppa_tools.lists.get_first_item:
            items:
              - Paris
              - London
              - Cardiff
              - Berlin
              - Brussels
          save: first_city
        ```

        The `first_city` variable will be a __string__ containing "Paris".
    """
    try:
        return items[0]
    except IndexError:
        message = "The supplied list is empty."
        ui.output_error(message)
        raise IndexError(message)


def get_item(items: List[Any], index: int) -> Any:
    """Get an item from a list using an index.

    !!! question "Minimum Plugin Version: 3.0.0"

    !!! tip "List Indexing"

        Values from a list are accessed using their position in the list (_index_).

        PPA lists are zero-indexed, meaning the first item is always at position `0`.

    ??? info "Input"

        - __items:__ a list of values

        - __index:__ the index of the item to get

    ??? success "Output"

        The value at position `index` in the `items` list.

    ??? example

        Getting the first item from a list of cities.

        ```YAML linenums="1"
        - ppa_tools.lists.get_item:
            items:
              - Paris
              - London
              - Cardiff
              - Berlin
              - Brussels
            index: 0
          save: selected_city
        ```

        The `selected_city` variable will be a __string__ containing "Paris".
    """
    return _get_item(items, index)


def get_items(items: List[Any], indices: List[int]) -> List[Any]:
    """Get items from a list using a list of indices.

    !!! question "Minimum Plugin Version: 3.0.0"

    !!! tip "List Indexing"

        Values from a list are accessed using their position in the list (_index_).

        PPA lists are zero-indexed, meaning the first item is always at position `0`.

    ??? info "Input"

        - __items:__ a list of values

        - __indices:__ a list of indices for the items to get

    ??? success "Output"

        A list of items taken from `items` at the positions supplied in `indices`.

    ??? example

        Getting indices __1__ & __3__ from a list of cities.

        ```YAML linenums="1"
        - ppa_tools.lists.get_items:
            items:
              - Paris
              - London
              - Cardiff
              - Berlin
              - Brussels
            indices:
              - 1
              - 3
          save: selected_cities
        ```

        The `selected_cities` variable will be a __list__ containing:

        ```YAML
        - London
        - Berlin
        ```
    """
    return [_get_item(items, index) for index in indices]
