"""
Timing

This module contains actions used to sleep & display countdowns in a PPA task.
"""
import uuid
import time

from typing import Optional

from ppa import ui


def sleep(seconds: int, progress_bar: Optional[bool] = False, start_text: Optional[str] = None, end_text: Optional[str] = None):
    """Wait a certain number of seconds before continuing the task.

    !!! question "Minimum Plugin Version: 1.1.0"

    ??? info "Input"

        - __seconds:__ the number of seconds to wait

        - __progress_bar:__ set to `true` to display a progress bar while the task waits

        - __start_text:__ the start text if _progress_bar_ is `true` (defaults to _The task is waiting for x seconds..._)

        - __end_text:__ the end text if _progress_bar_ is `true` (defaults to _The task will now continue_)

    ??? success "Output"
        Nothing is outputted by this action.

    ??? example

        ```YAML linenums="1"
        - ppa_tools.timing.sleep:
            seconds: 10
            progress_bar: true
            start_text: Pausing the task for 10 seconds...
            end_text: The 10 second pause is complete
        ```

    !!! danger "Remember"

        You must set _progress_bar_ to `true` for any supplied text to be displayed.
    """
    if progress_bar:
        element_id = str(uuid.uuid4())
        ui.output_progress(
            start_text or f"The task is waiting for {seconds} seconds...", percent=None, element_id=element_id,
        )
        time.sleep(seconds)
        ui.output_progress(end_text or "The task will now continue", percent=100, element_id=element_id)
    else:
        time.sleep(seconds)


def visual_countdown(seconds: int, doc: Optional[str] = "", refresh_interval: Optional[int] = 3) -> None:
    """Start a countdown displayed in a markdown document.

    The task will wait until the countdown is complete before moving to the next action.

    !!! question "Minimum Plugin Version: 1.1.0"

    ??? info "Input"

        - __doc:__ optional markdown to display with the countdown timer

        - __seconds:__ the number of seconds to count down from

        - __refresh_interval:__ the number of seconds between visual countdown refreshes (defaults to `3`)

    ??? success "Output"
        Nothing is outputted by this action.

    ??? example

        {% raw %}
        ```YAML linenums="1"
        - ppa_tools.timing.visual_countdown:
             doc: >
               {{ globals.icons.ppa.small }}

               This is a visual countdown timer...
             seconds: 10
             refresh_interval: 1
        ```
        {% endraw %}

    ??? note "Playbook Example Screenshot"

        <img src="../../../images/screenshots/ppa_tools/visual-countdown.png" alt="visual-countdown" width="100%"/>
    """
    countdown_text = "__Seconds Remaining:__ {remaining}"

    element_id = str(uuid.uuid4())
    markdown = "\n\n".join([doc, countdown_text])

    def _display(counter):
        if counter % refresh_interval == 0:
            ui.output_markdown(markdown.format(remaining=counter), element_id=element_id)
        time.sleep(1)

    [_display(i) for i in range(seconds, -1, -1)]
